# -*- coding: utf-8 -*-
# Copyright (c) 2010-2013 Infrae. All rights reserved.
# See also LICENSE.txt

import lxml
import lxml.cssselect
from collections import defaultdict
from collections import namedtuple

from infrae.testbrowser.form import Form
from infrae.testbrowser.utils import ResultSet
from infrae.testbrowser.utils import node_to_node, none_filter
from infrae.testbrowser.utils import resolve_url

def node_to_normalized_text(node):
    return ' '.join(
        [s for s in [s.strip() for s in node_to_text(node).split()] if s])

def node_to_text(node):
    return node.text_content().strip()

def tag_filter(name):
    def node_filter(node):
        return node.tag == name
    return node_filter


class Link(object):

    def __init__(self, html, browser):
        self.html = html
        self.browser = browser
        self.text = html.text_content().strip()

    @property
    def url(self):
        return resolve_url(self.html.attrib['href'], self.browser)

    def click(self):
        return self.browser.open(self.url)

    def __str__(self):
        if isinstance(self.text, str):
            return self.text.encode('utf-8', 'replace')
        return str(self.text)

    def __eq__(self, other):
        return self.text == other

    def __ne__(self, other):
        return self.text != other

    def __unicode__(self):
        return str(self.text)

    def __repr__(self):
        return repr(str(self))


class Links(ResultSet):

    def __init__(self, links, browser):
        super(Links, self).__init__(
            [(str(link).lower(), str(link), link) for link in [Link(link, browser) for link in links]])



ExpressionType = namedtuple(
    'ExpressionType',
    ('converter', 'filter', 'nodes', 'node'))

EXPRESSION_TYPE = {
    'node': ExpressionType(
        node_to_node,
        none_filter,
        lambda nodes, browser: list(nodes),
        lambda node, browser: node),
    'text': ExpressionType(
        node_to_text,
        none_filter,
        lambda nodes, browser: list(nodes),
        lambda node, browser: node),
    'normalized-text': ExpressionType(
        node_to_normalized_text,
        none_filter,
        lambda nodes, browser: list(nodes),
        lambda node, browser: node),
    'link': ExpressionType(
        node_to_node,
        tag_filter('a'),
        Links,
        Link),
    'form': ExpressionType(
        node_to_node,
        tag_filter('form'),
        lambda nodes, browser: [Form(n, browser) for n in nodes],
        Form),
    'form-fields': ExpressionType(
        node_to_node,
        tag_filter('form'),
        lambda nodes, browser: [Form(n, browser).inspect.fields for n in nodes],
        lambda node: Form(node).inspect.fields),
    'form-actions': ExpressionType(
        node_to_node,
        tag_filter('form'),
        lambda nodes, browser: [Form(n, browser).inspect.actions for n in nodes],
        lambda node: Form(node).inspect.actions),
    }


class Expressions(object):

    def __init__(self, browser):
        self.__browser = browser
        self.__expressions = defaultdict(lambda: tuple((None, None, None)))

    def add(self, name, xpath=None, type='text', css=None, unique=False):
        if type not in EXPRESSION_TYPE:
            raise AssertionError('Unknown expression type %s' % type)
        finder = None
        if xpath is not None:
            finder = lxml.etree.XPath(xpath)
        elif css is not None:
            finder = lxml.cssselect.CSSSelector(css)
        if finder is None:
            raise AssertionError(
                'You need to provide an XPath or CSS expression')
        self.__expressions[name] = (finder, type, unique)

    def __getattr__(self, name):
        finder, type, unique = self.__expressions[name]
        if finder is not None:
            if self.__browser.html is None:
                raise AssertionError('Not viewing HTML')
            expression = EXPRESSION_TYPE[type]
            nodes = list(filter(expression.filter,
                           list(map(expression.converter,
                               finder(self.__browser.html)))))
            if unique:
                if len(nodes) > 1:
                    raise AssertionError(
                        'Multiple elements found for %s where only '
                        'one was expected.' % name)
                if not len(nodes):
                    return None
                return expression.node(nodes[0], self.__browser)
            return expression.nodes(nodes, self.__browser)
        raise AttributeError(name)

