# -*- coding: utf-8 -*-
# Copyright (c) 2010-2013 Infrae. All rights reserved.
# See also LICENSE.txt

from zope.interface import Interface, Attribute

_marker = object()


class IBrowser(Interface):
    """Test browser interface.
    """
    inspect = Attribute("Browser Expression (inspect current HTML)")
    macros = Attribute("Browser macros")
    options = Attribute("Options")
    handlers = Attribute("Handlers used to customize browser behavior")
    cookies = Attribute("Handle access to the cookies")

    url = Attribute("Currently viewed URL")
    location = Attribute("Currently viewed path")
    contents = Attribute("Payload")
    html = Attribute("HTML payload parsed by LXML, or None")

    def __enter__():
        """Use a browser as a context manager. It will auto-close at
        the end of the context.
        """

    def __exit__(exc_type, exc_val, exc_tb):
        """Use a browser as a context manager. It will auto-close at
        the end of the context.
        """

    def open(url):
        """Open the given URL.
        """

    def close():
        """Close the browser.
        """

    def reload():
        """Reload the current opened URL, re-submitting [form] data if
        any where sent.
        """

    def login(user, password=_marker):
        """Set a HTTP Basic authorization header in the requests that
        are sent to the server. If no ``password`` is provided, the
        ``login`` will be used as ``password``.

        If ``login`` is None, it will remove any set HTTP Basic
        authorization header.
        """

    def logout():
        """Remove an HTTP Basic authorization set.
        """

    def get_form(name=None, id=None):
        """Return the identified form as a form object, or raise an
        AssertionError if no matching form are found.
        """

    def get_link(content):
        """Return a link that matches ``content``.
        """


class IAdvancedBrowser(IBrowser):
    """Browser with more low level API to the HTTP layer.
    """
    method = Attribute("Method used to view the current page")
    status = Attribute("HTTP status")
    status_code = Attribute("HTTP status code as an integer")
    content_type = Attribute("Content type")
    content_encoding = Attribute("Content encoding")
    headers = Attribute("Dictionary like access to response headers")
    history = Attribute("Last previously viewed URLs")
    xml = Attribute("XML payload parsed by LXML, or None")
    json = Attribute("JSON payload parsed, or None")

    def set_request_header(key, value):
        """Set an HTTP header ``key`` to the given ``value`` for each
        request made to the server.
        """

    def get_request_header(key):
        """Get the value or None corresponding to the HTTP header
        ``key`` that are sent to the server.
        """

    def del_request_header(key):
        """If set, remove the HTTP header ``key`` from each request
        made to the server.
        """

    def clear_request_headers():
        """Clear all custom added HTTP headers that are sent to the
        server each time a request is made.
        """

    def open(url, method='GET', query=None,
             form=None, form_enctype='application/x-www-form-urlencoded'):
        pass


# Cookie support


class ICookie(Interface):
    """Represent a single cookie.
    """
    name = Attribute("Name of the cookie")
    value = Attribute("Value of the cookie")
    options = Attribute("Dictionnary containing cookie options "
                        "(max-age, http-only)...")


class ICookies(Interface):
    """Represent all cookies currently set.
    """

    def add(name, value):
        """Add a new cookie called ``name`` with the given ``value``.
        """

    def clear():
        """Clear all set cookies.
        """

    def keys():
        """List all cookies names.
        """

    def __getitem__(key):
        """Get a cookie by its name.
        """

    def __contains__(key):
        """Return ``True`` if there is a cookie called ``key``,
        ``False`` otherwise.
        """

    def __len__():
        """Return the number cookies currently sets.
        """

    def __eq__(other):
        """Return ``True`` if the cookies name match the list
        ``other``, ``False`` otherwise.
        """

    def __ne__(other):
        """Return ``True`` if the cookies name doesn't match the list
        ``other``, ``False`` if they do match.
        """


# Form support

class IFormControl(Interface):
    """Represent a control in a form.
    """
    name = Attribute("Control name")
    type = Attribute("Control type (text, file, radio, select)")
    value = Attribute("Control value")
    multiple = Attribute("Does control hold multiple values?")
    options = Attribute("Vocabulary from which the control can "
                        "takes his values, if limited")
    checkable = Attribute("Is the control checkable?")
    checked = Attribute("If the control is checkable, is it checked")


class IClickableFormControl(IFormControl):

    def click():
        """Click that control.
        """


class ISubmitableFormControl(IClickableFormControl):

    def submit():
        """Submit the current form.
        """


class IForm(Interface):
    """Represent a form.
    """
    name = Attribute("HTML name of the form")
    action = Attribute("URL to which the form is submitted")
    method = Attribute("Method used to submit the form")
    enctype = Attribute("Encoding type used to submit the form")
    accept_charset = Attribute("Charset used to submit the form")
    controls = Attribute("Dict containing all the controls")

    def get_control(name):
        """Return the control identified by name or raise an
        AssertionError.
        """

    def submit(name=None, value=None):
        """Submit the form as-it. ``name`` and ``value`` let you add
        an extra value to the submission.
        """


# Browser options


class ICustomizableOptions(Interface):
    server = Attribute('Name of the server to use')
    port = Attribute('Port of the server to use')


class ISeleniumCustomizableOptions(Interface):
    # We don't inherit from ICustomizableOptions due zope.interface
    # API suckiness

    server = Attribute('Name of the server to use')
    port = Attribute('Port of the server to use')
    browser = Attribute('Browser to use firefox/internet explorer/chrome')
    selenium_host = Attribute('Host to use to connect to Selenium')
    selenium_port = Attribute('Port to use to connect to Selenium')
    selenium_platform = Attribute('Platform to request to Selenium')


# Advanced browser sub compoments.

class IWSGIResponse(Interface):
    status = Attribute('HTTP status line')
    headers = Attribute('Response headers')
    output = Attribute('Response data (except headers)')


class IWSGIServer(Interface):
    server = Attribute('Server hostname')
    port = Attribute('Server port')
    protocol = Attribute('HTTP procotol version')

    def get_default_environ():
        pass

    def get_environ(method, uri, headers, data=None, data_type=None):
        pass

    def __call__(method, uri, headers, data=None, data_type=None):
        """Compute the a response for the query. This return a
        IWSGIResponse object.
        """

